kB = 1.3806488e-23      # boltzmann constant
u = 1.66053892e-27      # atomic mass unit (proton mass)
h = 6.626070040e-34     # planck h
hbar = 1.054571800e-34  # planck hbar
c = 299792458           # speed of light
e = 1.602176565e-19     # electron charge
me = 9.10938291e-31     # electron mass
a0 = 5.291772109e-11    # bohr radius
muB = 9.274009994e-24   # bohr magneton
eps0 = 8.854187817e-12  # vacuum permittivity
